

#include "pmacros.h"

#include <stdio.h>
#include <sched.h>
#include <stdlib.h>
#include <semaphore.h>
#include <errno.h>

#include "returncode.h"


/*  Number of resources
*/
#define NUMBER_DISK 3
#define NUMBER_BUFFER 3


/*  Task states
*/
#define  WAIT 0
#define  DO_OTHER 1
#define  USE 2

/*  Number of maximum tasks
*/
#define NB_TASKS 10


/* Private semaphores
*/
sem_t sempriv[NB_TASKS];
pthread_mutex_t mutex;


/* Monitor data
*/
int task_state[NB_TASKS];
int task_disk_request[NB_TASKS];
int task_buffer_request[NB_TASKS];
int available_disk;
int available_buffer;


void allocate(int id, int disk_request, int buffer_request)
{
 int status;

 status=pthread_mutex_lock(&mutex);
 if (status!=0)
       returncode("mutex lock", status);


 if((available_disk>=disk_request) 
   && (available_buffer>=buffer_request))
 {
   status=sem_post(&sempriv[id]);
   if(status!=0)
	returncode("sem_post sempriv ",status);
   task_state[id]=USE;
   available_disk=available_disk-disk_request;
   available_buffer=available_buffer-buffer_request;
 }
 else
 {
   task_state[id]=WAIT;
   task_disk_request[id]=disk_request; 
   task_buffer_request[id]=buffer_request; 
 }

 status=pthread_mutex_unlock(&mutex);
 if (status!=0)
       returncode("mutex unlock", status);

 status=sem_wait(&sempriv[id]);
 if(status!=0)
	returncode("sem_wait sempriv ",status);

 printf("Thread number %d has allocated the resources\n",id);

}



void release(int id, int disk_request, int buffer_request)
{
 int status;
 int i;

 status=pthread_mutex_lock(&mutex);
 if (status!=0)
       returncode("mutex lock", status);

 task_state[id]=DO_OTHER;
 available_disk=available_disk+disk_request;
 available_buffer=available_buffer+buffer_request;

 for(i=0; i<NB_TASKS;i++) {
   if ((task_state[i] == WAIT) &&
       (available_disk>=disk_request) &&
       (available_buffer>=buffer_request))
     {
     task_state[i]=USE;
     available_disk=available_disk-disk_request;
     available_buffer=available_buffer-buffer_request;
     status=sem_post(&sempriv[i]);
     if(status!=0)
	returncode("sem_post sempriv ",status);
     }
 }

 status=pthread_mutex_unlock(&mutex);
 if (status!=0)
       returncode("mutex unlock", status);

 printf("Thread number %d has released the resources\n",id);

}



void init_monitor(void)
{
 int status;
 int i;

 available_disk=NUMBER_DISK;
 available_buffer=NUMBER_BUFFER;

 status=pthread_mutex_init(&mutex, NULL);
 if(status!=0)
	returncode("pthread_mutex_init ",status);


 for(i=0; i<NB_TASKS;i++) {
   task_state[i]=DO_OTHER;
   task_disk_request[i]=0;
   task_buffer_request[i]=0;
   status=sem_init(&sempriv[i],0,0);
   if(status!=0)
	returncode("sem_init sem_priv",status);
 }

}

